@@ -16,6 +16,8 @@
 from enum import IntFlag, auto, Enum
 
 
+ERROR = -1
+
 DEF_SAVE_DIR = 'ww-img'
 
 LAYERS = 'layers'
@@ -26,7 +28,11 @@ FULL_SVD = 'full_svd'
 
 # fi_ fingers options
 XMIN_PEAK = 'xmin_peak'
+XMIN_MAX = 'xmin_max'
 CLIP_XMAX = 'clip_xmax'
+MAX_N = 'max_N'
+DEFAULT_MAX_N = 10
+
 
 SVD = 'svd' # TruncatedSVD for Smoothing
 RMT = 'rmt' # pyRMT / RIE Removed for 0.6.5
@@ -39,6 +45,11 @@ POWER_LAW = 'power_law'
 LOG_NORMAL = 'lognormal'
 EXPONENTIAL = 'exponential'
 
+# distance choices
+EUCLIDEAN = 'euclidean'
+CKA = 'cka'
+RAW = 'raw'
+
 # STATUSes
 
 SUCCESS = 'success'
@@ -83,6 +94,7 @@ MIN_NUM_EVALS = 10
 
 MAX_EVALS = 'max_evals'
 DEFAULT_MAX_EVALS = 10000
+MAX_NUM_EVALS= DEFAULT_MAX_EVALS
 
 PLOT = 'plot'
 STACKED = 'stacked'
@@ -93,6 +105,11 @@ LAST = 'last'
 
 TOLERANCE = 'tolerance'
 WEAK_RANK_LOSS_TOLERANCE = 0.000001 # on ei=gen values
+
+
+ADD_BIASES = 'add_biases'
+
+DEFAULT_START_ID = 0
     
 # These are NOT the defaults...see analyze() for actual defaults
 DEFAULT_PARAMS = {GLOROT_FIX: False, NORMALIZE:False, CONV2D_NORM:True, RANDOMIZE: True, 
@@ -101,11 +118,12 @@ DEFAULT_PARAMS = {GLOROT_FIX: False, NORMALIZE:False, CONV2D_NORM:True, RANDOMIZ
                   WW2X:False, VECTORS:True, SMOOTH:None, STACKED:False, 
                   SVD_METHOD:FULL_SVD,  FIX_FINGERS:None, FIT:POWER_LAW, 
                   SPARSIFY: True, DETX: True,  MP_FIT:False,
-                  MIN_EVALS:DEFAULT_MIN_EVALS, MAX_EVALS:DEFAULT_MAX_EVALS, 
-                  TOLERANCE:WEAK_RANK_LOSS_TOLERANCE}
+                  MIN_EVALS:DEFAULT_MIN_EVALS, MAX_EVALS:DEFAULT_MAX_EVALS, MAX_N:DEFAULT_MAX_N,
+                  TOLERANCE:WEAK_RANK_LOSS_TOLERANCE, START_IDS:DEFAULT_START_ID, ADD_BIASES:False}
 
 
 EVALS_THRESH = 0.00001
+ALMOST_ZERO = 1e-10
 
 class LAYER_TYPE(IntFlag):
     UNKNOWN = auto()
@@ -117,12 +135,18 @@ class LAYER_TYPE(IntFlag):
     EMBEDDING = auto()
     NORM = auto()
 
-    
+# framework names
+KERAS = 'kers'
+PYTORCH = 'pytorch'
+PYSTATEDICT = 'pystatedict'
+ONNX = 'onnx'
+
 class FRAMEWORK(IntFlag):
     UNKNOWN = auto()
     PYTORCH = auto()
     KERAS = auto()
     ONNX = auto()
+    PYSTATEDICT = auto()
     
 
 class CHANNELS(IntFlag):
@@ -130,7 +154,8 @@ class CHANNELS(IntFlag):
     FIRST = auto()
     LAST = auto()   
             
-    
+            
+# TODO either complete or remove thi 
 class METRICS():
     NORM = "norm"
     LOG_NORM = "log_norm"
